#!/usr/bin/env node
"use strict";
/* --------------------------------------------------------------------------------------------
 * Copyright (c) Eugen Wiens. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.connection = void 0;
const path_1 = __importDefault(require("path"));
const node_1 = require("vscode-languageserver/node");
const BitBakeDocScanner_1 = require("./BitBakeDocScanner");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const analyzer_1 = require("./tree-sitter/analyzer");
const parser_1 = require("./tree-sitter/parser");
const OutputLogger_1 = require("./lib/src/utils/OutputLogger");
const onCompletion_1 = require("./connectionHandlers/onCompletion");
const onDefinition_1 = require("./connectionHandlers/onDefinition");
const onHover_1 = require("./connectionHandlers/onHover");
const general_support_1 = require("./embedded-languages/general-support");
const semanticTokens_1 = require("./semanticTokens");
const BitbakeProjectScannerClient_1 = require("./BitbakeProjectScannerClient");
const requests_1 = require("./lib/src/types/requests");
const notifications_1 = require("./lib/src/types/notifications");
const BitbakeSettings_1 = require("./lib/src/BitbakeSettings");
const files_1 = require("./lib/src/utils/files");
const onReference_1 = require("./connectionHandlers/onReference");
// Create a connection for the server. The connection uses Node's IPC as a transport
exports.connection = (0, node_1.createConnection)(node_1.ProposedFeatures.all);
(0, onDefinition_1.setDefinitionsConnection)(exports.connection);
const documents = new node_1.TextDocuments(vscode_languageserver_textdocument_1.TextDocument);
let parseOnSave = true;
let workspaceFolder;
let pokyFolder;
const disposables = [];
let currentActiveTextDocument = vscode_languageserver_textdocument_1.TextDocument.create('file://dummy_uri', 'bitbake', 0, '');
exports.connection.onInitialize(async (params) => {
    OutputLogger_1.logger.level = params.initializationOptions?.loggingLevel ?? 'error';
    OutputLogger_1.logger.info('[onInitialize] Initializing connection');
    BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.setConnection(exports.connection);
    disposables.push(...BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.buildHandlers());
    workspaceFolder = params.workspaceFolders?.[0].uri.replace('file://', '');
    // If the language server is not started by VSCode, extensionPath and pathToBitbakeFolder can be undefined
    // We provide alternatives here to prevent crashes when these variables are used
    const extensionPath = params.initializationOptions?.extensionPath ?? path_1.default.join(__dirname, '..');
    pokyFolder = pokyFolder ?? workspaceFolder;
    OutputLogger_1.logger.info('[onInitialize] Setting yocto doc path and parsing doc files');
    BitBakeDocScanner_1.bitBakeDocScanner.setDocPathAndParse(extensionPath);
    const parser = await (0, parser_1.generateParser)();
    analyzer_1.analyzer.initialize(parser);
    BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.onChange.on('scanReady', () => {
        OutputLogger_1.logger.debug('Analyzing the current document again...');
        analyzer_1.analyzer.analyze({ document: currentActiveTextDocument, uri: currentActiveTextDocument.uri });
    });
    return {
        capabilities: {
            textDocumentSync: node_1.TextDocumentSyncKind.Incremental,
            completionProvider: {
                resolveProvider: true,
                triggerCharacters: [':', '[']
            },
            definitionProvider: true,
            referencesProvider: true,
            hoverProvider: true,
            semanticTokensProvider: {
                legend: semanticTokens_1.legend,
                full: true
            }
        }
    };
});
exports.connection.onShutdown(() => {
    disposables.forEach((disposable) => { disposable.dispose(); });
});
exports.connection.onDidChangeConfiguration((change) => {
    OutputLogger_1.logger.level = change.settings.bitbake?.loggingLevel ?? OutputLogger_1.logger.level;
    parseOnSave = change.settings.bitbake?.parseOnSave ?? parseOnSave;
    const bitbakeFolder = (0, BitbakeSettings_1.expandSettingPath)(change.settings.bitbake?.pathToBitbakeFolder, { workspaceFolder });
    if (bitbakeFolder !== undefined) {
        pokyFolder = path_1.default.join(bitbakeFolder, '..'); // We assume BitBake is into Poky
    }
});
exports.connection.onCompletion(onCompletion_1.onCompletionHandler);
exports.connection.onCompletionResolve((item) => {
    OutputLogger_1.logger.debug(`[onCompletionResolve]: ${JSON.stringify(item)}`);
    return item;
});
exports.connection.onDefinition(onDefinition_1.onDefinitionHandler);
exports.connection.onReferences(onReference_1.onReferenceHandler);
exports.connection.onHover(onHover_1.onHoverHandler);
exports.connection.onRequest(requests_1.RequestMethod.EmbeddedLanguageTypeOnPosition, async ({ uriString, position }) => {
    return (0, general_support_1.getEmbeddedLanguageTypeOnPosition)(uriString, position);
});
// This request method 'textDocument/semanticTokens' will be sent when semanticTokensProvider capability is enabled
exports.connection.onRequest('textDocument/semanticTokens/full', ({ textDocument }) => {
    OutputLogger_1.logger.debug(`[OnRequest] <textDocument/semanticTokens/full> Document uri: ${textDocument.uri}`);
    return (0, semanticTokens_1.getSemanticTokens)(textDocument.uri);
});
exports.connection.onRequest(requests_1.RequestMethod.getLinksInDocument, (params) => {
    return analyzer_1.analyzer.getLinksInStringContent(params.documentUri);
});
exports.connection.onRequest(requests_1.RequestMethod.ProcessRecipeScanResults, (param) => {
    OutputLogger_1.logger.debug(`[onNotification] <ProcessRecipeScanResults> uri:  ${JSON.stringify(param.uri)} recipe: ${param.chosenRecipe}`);
    analyzer_1.analyzer.processRecipeScanResults(param.scanResults, param.chosenRecipe);
});
exports.connection.onRequest('bitbake/getVar', async (params) => {
    const scanResult = analyzer_1.analyzer.getLastScanResult(params.recipe);
    return scanResult?.symbols.find(symbolInfo => symbolInfo.name === params.variable)?.finalValue;
});
exports.connection.onNotification(notifications_1.NotificationMethod.RemoveScanResult, (param) => {
    OutputLogger_1.logger.debug(`[onNotification] <${notifications_1.NotificationMethod.RemoveScanResult}> recipe: ${param.recipeName}`);
    analyzer_1.analyzer.removeLastScanResultForRecipe(param.recipeName);
});
exports.connection.listen();
const analyzeDocument = async (event) => {
    const textDocument = event.document;
    const previousVersion = analyzer_1.analyzer.getAnalyzedDocument(textDocument.uri)?.version ?? -1;
    if (textDocument.getText().length > 0 && previousVersion < textDocument.version) {
        const diagnostics = analyzer_1.analyzer.analyze({ document: textDocument, uri: textDocument.uri });
        const embeddedLanguageDocs = (0, general_support_1.generateEmbeddedLanguageDocs)(event.document, pokyFolder);
        if (embeddedLanguageDocs !== undefined) {
            void exports.connection.sendNotification(notifications_1.NotificationMethod.EmbeddedLanguageDocs, embeddedLanguageDocs);
        }
        void exports.connection.sendDiagnostics({ uri: textDocument.uri, diagnostics });
    }
    currentActiveTextDocument = textDocument;
    // Other language extensions might also associate .conf files with their langauge modes
    if (textDocument.uri.endsWith('.conf')) {
        // get decoded urls in case of running on Windows
        const filePath = decodeURIComponent(new URL(textDocument.uri).pathname);
        OutputLogger_1.logger.debug(`[Verify Configuration File Association] file uri: ${filePath}`);
        await exports.connection.sendRequest('bitbake/verifyConfigurationFileAssociation', { filePath });
    }
};
documents.onDidOpen(analyzeDocument);
documents.onDidChangeContent(analyzeDocument);
documents.onDidSave(async (event) => {
    OutputLogger_1.logger.info(`[onDidSave] Document saved: ${event.document.uri}`);
    if (parseOnSave) {
        const exts = ['.bb', '.bbappend', '.inc'];
        const uri = event.document.uri;
        if (exts.includes(path_1.default.extname(uri))) {
            const foundRecipe = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._recipes.find((recipe) => recipe.name === (0, files_1.extractRecipeName)(uri));
            if (foundRecipe !== undefined) {
                OutputLogger_1.logger.debug(`[onDidSave] Running 'bitbake -e' against the saved recipe: ${foundRecipe.name}`);
                // Note that it pends only one scan at a time. See client/src/driver/BitbakeRecipeScanner.ts.
                // Saving more than 2 files at the same time could cause the server to miss some of the scans.
                void exports.connection.sendRequest('bitbake/scanRecipe', { uri });
                return;
            }
        }
        // saving other files or no recipe is resolved
        void exports.connection.sendRequest('bitbake/parseAllRecipes');
    }
});
documents.listen(exports.connection);
//# sourceMappingURL=server.js.map