"use strict";
/* --------------------------------------------------------------------------------------------
 * Copyright (c) 2023 Savoir-faire Linux. All rights reserved.
 * Licensed under the MIT License. See License.txt in the project root for license information.
 * ------------------------------------------------------------------------------------------ */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyzer = void 0;
/**
 * Inspired by bash-language-server under MIT
 * Reference: https://github.com/bash-lsp/bash-language-server/blob/8c42218c77a9451b308839f9a754abde901323d5/server/src/analyser.ts
 */
const vscode_languageserver_1 = require("vscode-languageserver");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const declarations_1 = require("./declarations");
const TreeSitterUtils = __importStar(require("./utils"));
const OutputLogger_1 = require("../lib/src/utils/OutputLogger");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const BitbakeProjectScannerClient_1 = require("../BitbakeProjectScannerClient");
const BitBakeDocScanner_1 = require("../BitBakeDocScanner");
class Analyzer {
    constructor() {
        this.uriToAnalyzedDocument = {};
        this.uriToLastScanResult = {}; // Store the results of the last scan for each recipe
    }
    getDocumentTexts(uri) {
        return this.uriToAnalyzedDocument[uri]?.document.getText().split(/\r?\n/g);
    }
    getAnalyzedDocument(uri) {
        return this.uriToAnalyzedDocument[uri];
    }
    getLastScanResult(recipe) {
        return this.uriToLastScanResult[recipe];
    }
    getIncludeUrisForUri(uri) {
        return this.uriToAnalyzedDocument[uri]?.includeFileUris ?? [];
    }
    getVariableExpansionSymbols(uri) {
        return this.uriToAnalyzedDocument[uri]?.variableExpansionSymbols ?? [];
    }
    removeLastScanResultForRecipe(recipe) {
        this.uriToLastScanResult[recipe] = undefined;
    }
    initialize(parser) {
        this.parser = parser;
    }
    analyze({ document, uri }) {
        if (this.parser === undefined) {
            OutputLogger_1.logger.debug('[Analyzer] The analyzer is not initialized with a parser');
            return [];
        }
        const fileContent = document.getText();
        const tree = this.parser.parse(fileContent);
        const globalDeclarations = (0, declarations_1.getGlobalDeclarations)({ tree, uri });
        const variableExpansionSymbols = this.getVariableExpansionSymbolsFromTree({ tree, uri });
        this.uriToAnalyzedDocument[uri] = {
            version: document.version,
            document,
            globalDeclarations,
            variableExpansionSymbols,
            includeFileUris: this.extractIncludeFileUris(uri, tree),
            tree
        };
        return this.executeAnalyzation(document, uri, tree);
    }
    executeAnalyzation(document, uri, tree) {
        const diagnostics = [];
        // It was used to provide diagnostics from tree-sitter, but it is not yet reliable.
        return diagnostics;
    }
    getVariableExpansionSymbolsFromTree({ tree, uri }) {
        const variableExpansionSymbols = [];
        TreeSitterUtils.forEach(tree.rootNode, (node) => {
            const isNonEmptyVariableExpansion = (node.type === 'identifier' && node.parent?.type === 'variable_expansion');
            const followChildren = !isNonEmptyVariableExpansion;
            if (isNonEmptyVariableExpansion) {
                const symbol = (0, declarations_1.nodeToSymbolInformation)({ node, uri, getFinalValue: false, isVariableExpansion: true });
                symbol !== null && variableExpansionSymbols.push(symbol);
            }
            return followChildren;
        });
        return variableExpansionSymbols;
    }
    getGlobalDeclarationSymbols(uri) {
        const analyzedDocument = this.uriToAnalyzedDocument[uri];
        if (analyzedDocument !== undefined) {
            const { globalDeclarations } = analyzedDocument;
            return this.getAllSymbolsFromGlobalDeclarations(globalDeclarations);
        }
        return [];
    }
    getAllSymbolsFromGlobalDeclarations(globalDeclarations) {
        let symbols = [];
        Object.values(globalDeclarations).forEach((symbolArray) => {
            symbols = symbols.concat(symbolArray);
        });
        return symbols;
    }
    findExactSymbolAtPoint(uri, position, wordAtPoint) {
        return exports.analyzer.getGlobalDeclarationSymbols(uri).find((symbol) => symbol.name === wordAtPoint && exports.analyzer.positionIsInRange(position.line, position.character, symbol.location.range)) ?? exports.analyzer.uriToAnalyzedDocument[uri]?.variableExpansionSymbols.find((symbol) => symbol.name === wordAtPoint && exports.analyzer.positionIsInRange(position.line, position.character, symbol.location.range));
    }
    getParsedTreeForUri(uri) {
        return this.uriToAnalyzedDocument[uri]?.tree;
    }
    /**
     * Find the full word at the given point.
     */
    wordAtPoint(uri, line, column) {
        const node = this.nodeAtPoint(uri, line, column);
        if (node === null || node.childCount > 0 || node.text.trim() === '') {
            return null;
        }
        return node.text.trim();
    }
    wordAtPointFromTextPosition(params) {
        return this.wordAtPoint(params.textDocument.uri, params.position.line, params.position.character);
    }
    hasParser() {
        return this.parser !== undefined;
    }
    resetAnalyzedDocuments() {
        this.uriToAnalyzedDocument = {};
    }
    /**
     * Get the directive keyword when the statement is either include, require or inherit directive by looking up the tree nodes
     */
    getDirectiveStatementKeywordByNodeType(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        const type = n?.type;
        const parentType = n?.parent?.type;
        switch (type) {
            case 'inherit_path':
                if (parentType === 'inherit_directive') {
                    return 'inherit';
                }
                return undefined;
            case 'include_path':
                if (parentType === 'require_directive') {
                    return 'require';
                }
                else if (parentType === 'include_directive') {
                    return 'include';
                }
                return undefined;
            default:
                return undefined;
        }
    }
    getDirectivePathForPosition(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        if (n?.type === 'inherit_path' || n?.type === 'include_path') {
            return n.text;
        }
        return undefined;
    }
    isIdentifier(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        return n?.type === 'identifier';
    }
    isFunctionIdentifier(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        if (n?.type === 'identifier') {
            return n?.parent?.type === 'function_definition' || n?.parent?.type === 'anonymous_python_function';
        }
        else if (n?.type === 'python_identifier') {
            return n?.parent?.type === 'python_function_definition';
        }
        return false;
    }
    isStringContent(uri, line, column) {
        const n = this.nodeAtPoint(uri, line, column);
        if (n?.type === 'string_content') {
            return true;
        }
        return false;
    }
    isOverride(uri, line, column) {
        const n = this.nodeAtPoint(uri, line, column);
        // Current tree-sitter (as of @1.0.1) only treats 'append', 'prepend' and 'remove' as a node with "override" type. Other words following ":" will yield a node with "identifier" type whose parent node is of type "override"
        // However, in some cases, its parent node is not of type override but its grandparent is.
        // Some ugly checks were added (as of tree-sitter @1.1.0) due to: https://github.com/amaanq/tree-sitter-bitbake/issues/9
        // See if future tree-sitter has a nicer way to handle this.
        return n?.type === 'override' ||
            n?.parent?.type === 'override' ||
            n?.parent?.parent?.type === 'override' ||
            (n?.type === ':' && n?.parent?.firstNamedChild?.type === 'identifier') || // when having something like "MYVAR:append:" at the last line of the document
            (n?.type === ':' && n?.parent?.type === 'ERROR' && n?.parent?.previousSibling?.type === 'override'); // when having MYVAR:append: = '123' and the second or later colon is typed
    }
    // While the user is typing on a line right before a function, tree-sitter-bitbake wrongly assumes the identifier is part of the function.
    // See tree-sitter-bitbake issue: https://github.com/amaanq/tree-sitter-bitbake/issues/16
    // TODO: Remove this function when the issue is fixed
    isBuggyIdentifier(uri, line, column) {
        const n = this.nodeAtPoint(uri, line, column);
        // The bug appears on identifiers, and also on the colon of overrides
        if (n?.type !== 'identifier' && n?.type !== ':') {
            return false;
        }
        const currentLine = n.startPosition.row;
        const parent = n.parent;
        // The bug occurs when tree-sitter-bitbake mistakens the identifier as part of the name of a function
        // The name of a function is on the same line as a function_definition node
        if (parent?.type !== 'function_definition' && currentLine !== parent?.startPosition.row) {
            return false;
        }
        const nextSibling = n.nextSibling;
        // If the identifier is the name of a function, then the next sibling should be '(' or an override.
        // In any case, it should be on the same line as the identifier.
        // If it is not, then we have our buggy identifier.
        return nextSibling?.endPosition.row !== currentLine;
    }
    isInsideBashRegion(uri, line, column) {
        let n = this.nodeAtPoint(uri, line, column);
        if (this.isBuggyIdentifier(uri, line, column)) {
            return false;
        }
        while (n !== null) {
            if (TreeSitterUtils.isShellDefinition(n)) {
                return true;
            }
            n = n.parent;
        }
        return false;
    }
    isInsidePythonRegion(uri, line, column) {
        let n = this.nodeAtPoint(uri, line, column);
        if (this.isBuggyIdentifier(uri, line, column)) {
            return false;
        }
        while (n !== null) {
            if (TreeSitterUtils.isInlinePython(n) || TreeSitterUtils.isPythonDefinition(n)) {
                return true;
            }
            n = n.parent;
        }
        return false;
    }
    isPythonDatastoreVariable(uri, line, column, 
    // Whether or not the opening quote should be considered as part of the variable
    // We want to be able to suggest completion items as the user open the quotes
    // However, when the user hover on the variable, the quotes are not part of the variable
    includeOpeningQuote = false) {
        const n = this.nodeAtPoint(uri, line, column);
        if (!this.isInsidePythonRegion(uri, line, column)) {
            return false;
        }
        if (n?.type !== 'string_content' && (includeOpeningQuote ? n?.type !== 'string_start' : true)) {
            return false;
        }
        // Example:
        // n.text: FOO
        // n.parent.text: 'FOO'
        // n.parent.parent.text: ('FOO')
        // n.parent.parent.parent.text: d.getVar('FOO')
        const parentParentParent = n?.parent?.parent?.parent;
        if (parentParentParent?.type !== 'call') {
            return false;
        }
        const match = parentParentParent.text.match(/^(d|e\.data)\.(?<name>.*)\((?<params>.*)\)$/); // d.name(params), e.data.name(params)
        const functionName = match?.groups?.name;
        if (functionName === undefined || !(BitBakeDocScanner_1.bitBakeDocScanner.pythonDatastoreFunction.includes(functionName))) {
            return false;
        }
        // Example for d.getVar('FOO'):
        // n.text: FOO
        // n.parent.text: 'FOO'
        // n.parent.previousSibling.text: (
        const isFirstParameter = n?.parent?.previousSibling?.text === '(';
        if (isFirstParameter) {
            const firstParameter = match?.groups?.params?.split(',')[0]?.trim();
            return firstParameter === n?.parent?.text;
        }
        // Example for d.renameVar('FOO', 'BAR'):
        // n.text: BAR
        // n.parent.text: 'BAR'
        // n.parent.previousSibling.text: ,
        // n.parent.previousSibling.previousSibling.text: 'FOO'
        // n.parent.previousSibling.previousSibling.previousSibling.text: (
        const isSecondParameter = n?.parent?.previousSibling?.previousSibling?.previousSibling?.text === '(';
        // d.renameVar is the only function for which the second parameter could be a Yocto defined variable
        if (functionName === 'renameVar' && isSecondParameter) {
            const secondParameter = match?.groups?.params?.split(',')[1]?.trim();
            return secondParameter === n?.parent?.text;
        }
        return false;
    }
    /**
     * Check if the variable expansion syntax is being typed. Only for expressions that reference variables. \
     * Example:
     * ```
     * NAME = "foo"
     * DESCRIPTION = "Name: ${NAME}"
     * ```
     */
    isVariableExpansion(uri, line, column) {
        const n = this.nodeAtPoint(uri, line, column);
        // since @1.0.2 the tree-sitter gives empty variable expansion (e.g. `VAR = "${}""`) the type "variable_expansion". But the node type returned from nodeAtPoint() at the position between "${" and "}" is of type "${" which is the result from descendantForPosition() (It returns the smallest node containing the given postion). In this case, the parent node has type "variable_expansion". Hence, we have n?.parent?.type === 'variable_expansion' below. The second expression after the || will be true if it encounters non-empty variable expansion syntax. e.g. `VAR = "${A}". Note that inline python with ${@} has type "inline_python"
        return n?.parent?.type === 'variable_expansion' || (n?.type === 'identifier' && n?.parent?.type === 'variable_expansion');
    }
    /**
     * Check if the node is the identifier in a variable assignment syntax (identifiers are only on the left hand side)
     */
    isIdentifierOfVariableAssignment(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        return n?.type === 'identifier' && n?.parent?.type === 'variable_assignment';
    }
    isVariableFlag(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        return n?.type === 'flag' && n?.parent?.type === 'variable_flag';
    }
    rangeForWordAtPoint(params) {
        const n = this.nodeAtPoint(params.textDocument.uri, params.position.line, params.position.character);
        if (n === null || n.childCount > 0 || n.text.trim() === '') {
            return undefined;
        }
        return TreeSitterUtils.range(n);
    }
    /**
     * Check if the current line starts with any directive statement keyword. The keyword is one of `include`, `inherit` and `require`
     *
     * Tree-sitter functionalities are not used here since they (as of @1.0.1) can't reliably treat a line as directive statement if the keyword presents but nothing follows.
     */
    getDirectiveStatementKeywordByLine(textDocumentPositionParams) {
        const { textDocument, position } = textDocumentPositionParams;
        const documentAsText = this.getDocumentTexts(textDocument.uri);
        if (documentAsText === undefined) {
            return undefined;
        }
        const currentLine = documentAsText[position.line];
        const lineTillCurrentPosition = currentLine.substring(0, position.character);
        const words = lineTillCurrentPosition.split(' ');
        if (words[0] === 'include' || words[0] === 'inherit' || words[0] === 'require') {
            return words[0];
        }
        return undefined;
    }
    /**
     * Get the keyword based on the node type in the tree
     */
    getKeywordForPosition(uri, line, column) {
        const n = this.nodeAtPoint(uri, line, column);
        const parentNodeType = n?.parent?.type;
        if (parentNodeType !== undefined) {
            switch (parentNodeType) {
                case 'inherit_directive':
                    return 'inherit';
                case 'include_directive':
                    return 'include';
                case 'require_directive':
                    return 'require';
                // other keywords...
                default:
                    return undefined;
            }
        }
        return undefined;
    }
    /**
     * Find the node at the given point.
     */
    nodeAtPoint(uri, line, column) {
        const tree = this.uriToAnalyzedDocument[uri]?.tree;
        if (tree === undefined) {
            return null;
        }
        if (tree.rootNode === null) {
            // Check for lacking rootNode (due to failed parse?)
            return null;
        }
        return tree.rootNode.descendantForPosition({ row: line, column });
    }
    // Return the uris in the diretive statements for unlimited depth
    extractIncludeFileUris(uri, tree) {
        const includeUris = [];
        this.sourceIncludeFiles(uri, includeUris, tree);
        return includeUris;
    }
    /**
     * The files pointed by the include URIs will analyzed if not yet done so such that the symbols in the included files are available for querying.
     */
    sourceIncludeFiles(uri, includeFileUris, tree) {
        if (this.parser === undefined) {
            OutputLogger_1.logger.error('[Analyzer] The analyzer is not initialized with a parser');
            return;
        }
        const filePath = uri.replace('file://', '');
        OutputLogger_1.logger.debug(`[Analyzer] Sourcing file: ${filePath}`);
        try {
            let parsedTree;
            if (tree !== undefined) {
                parsedTree = tree;
            }
            else {
                const analyzedDocument = this.uriToAnalyzedDocument[uri];
                if (analyzedDocument === undefined) {
                    const textDocument = vscode_languageserver_textdocument_1.TextDocument.create(uri, 'bitbake', 0, fs_1.default.readFileSync(uri.replace('file://', ''), 'utf8'));
                    parsedTree = this.parser.parse(textDocument.getText());
                    // Store it in analyzedDocument just like what analyze() does to avoid re-reading the file from disk and re-parsing the tree when editing on the same file
                    this.uriToAnalyzedDocument[uri] = {
                        version: textDocument.version,
                        document: textDocument,
                        globalDeclarations: (0, declarations_1.getGlobalDeclarations)({ tree: parsedTree, uri }),
                        variableExpansionSymbols: this.getVariableExpansionSymbolsFromTree({ tree: parsedTree, uri }),
                        includeFileUris: [],
                        tree: parsedTree
                    };
                }
                else {
                    OutputLogger_1.logger.debug('[Analyzer] File already analyzed');
                    parsedTree = analyzedDocument.tree;
                }
            }
            // Recursively scan for files in the directive statements `inherit`, `include` and `require` and pass the same reference of extraSymbols to each recursive call
            const fileUris = this.getDirectiveFileUris(parsedTree);
            fileUris.forEach(uri => {
                if (!includeFileUris.includes(uri)) {
                    includeFileUris.push(uri);
                }
            });
            for (const fileUri of fileUris) {
                this.sourceIncludeFiles(fileUri, includeFileUris);
            }
        }
        catch (error) {
            if (error instanceof Error) {
                OutputLogger_1.logger.error(`[Analyzer] Error reading file at ${filePath}: ${error.message}`);
            }
            else if (typeof error === 'string') {
                OutputLogger_1.logger.error(`[Analyzer] Error reading file at ${filePath}: ${error}`);
            }
            else {
                OutputLogger_1.logger.error(`[Analyzer] An unknown error occurred while reading the file at ${filePath}`);
            }
        }
    }
    getDirectiveFileUris(parsedTree) {
        const fileUris = [];
        parsedTree.rootNode.children.forEach((childNode) => {
            if (childNode.type === 'inherit_directive') {
                childNode.children.forEach((n) => {
                    if (n.type === 'inherit_path') {
                        OutputLogger_1.logger.debug(`[Analyzer] Found inherit path: ${n.text}`);
                        const bbclasses = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._classes.filter((bbclass) => {
                            return bbclass.name === n.text;
                        });
                        for (const bbclass of bbclasses) {
                            if (bbclass.path !== undefined) {
                                const uri = 'file://' + bbclass.path.dir + '/' + bbclass.path.base;
                                fileUris.push(encodeURI(uri));
                            }
                        }
                    }
                });
            }
            else if (childNode.type === 'require_directive' || childNode.type === 'include_directive') {
                if (childNode.firstNamedChild !== null && childNode.firstNamedChild.type === 'include_path') {
                    OutputLogger_1.logger.debug(`[Analyzer] Found path: ${childNode.firstNamedChild.text}`);
                    const foundFiles = this.findFilesInProjectScanner(childNode.firstNamedChild.text);
                    for (const file of foundFiles) {
                        if (file.path !== undefined) {
                            const uri = 'file://' + file.path.dir + '/' + file.path.base;
                            fileUris.push(encodeURI(uri));
                        }
                    }
                }
            }
        });
        return fileUris;
    }
    findFilesInProjectScanner(filePath) {
        const parsedPath = path_1.default.parse(filePath);
        let foundFiles = [];
        switch (parsedPath.ext) {
            case '.inc':
                foundFiles = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._includes.filter((inc) => {
                    return inc.name === parsedPath.name;
                });
                break;
            case '.bb':
                foundFiles = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._recipes.filter((recipe) => {
                    return recipe.name === parsedPath.name;
                });
                break;
            case '.conf':
                foundFiles = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._confFiles.filter((conf) => {
                    return conf.name === parsedPath.name;
                });
                break;
            default:
                OutputLogger_1.logger.warn(`[Analyzer] Unsupported file extension: ${parsedPath.ext}`);
                break;
        }
        return foundFiles;
    }
    /**
     * Extract symbols from the string content of the tree
     */
    getSymbolsInStringContent(uri, line, character) {
        const allSymbolsAtPosition = [];
        const wholeWordRegex = /(?<![-.:])(--(enable|disable)-)?\b(?<name>[a-zA-Z0-9][a-zA-Z0-9-+.]*[a-zA-Z0-9])\b(?![-.:])/g;
        const n = this.nodeAtPoint(uri, line, character);
        if (n?.type === 'string_content') {
            this.processSymbolsInStringContent(n, wholeWordRegex, (start, end, match) => {
                const symbolName = match.groups?.name;
                if (symbolName !== undefined) {
                    OutputLogger_1.logger.debug(`[Analyzer] Found symbol in string content: ${symbolName}`);
                    if (this.positionIsInRange(line, character, { start, end })) {
                        const foundRecipe = BitbakeProjectScannerClient_1.bitBakeProjectScannerClient.bitbakeScanResult._recipes.find((recipe) => {
                            return recipe.name === symbolName;
                        });
                        if (foundRecipe !== undefined) {
                            if (foundRecipe?.path !== undefined) {
                                allSymbolsAtPosition.push({
                                    name: symbolName,
                                    kind: vscode_languageserver_1.SymbolKind.Variable,
                                    location: {
                                        range: {
                                            start,
                                            end
                                        },
                                        uri: 'file://' + foundRecipe.path.dir + '/' + foundRecipe.path.base
                                    }
                                });
                            }
                            if (foundRecipe?.appends !== undefined && foundRecipe.appends.length > 0) {
                                foundRecipe.appends.forEach((append) => {
                                    allSymbolsAtPosition.push({
                                        name: append.name,
                                        kind: vscode_languageserver_1.SymbolKind.Variable,
                                        location: {
                                            range: {
                                                start,
                                                end
                                            },
                                            uri: 'file://' + append.dir + '/' + append.base
                                        }
                                    });
                                });
                            }
                        }
                    }
                }
            });
        }
        return allSymbolsAtPosition;
    }
    getLinksInStringContent(uri) {
        const links = [];
        const uriRegex = /(file:\/\/)(?<uri>.*)\b/g;
        const parsedTree = this.getAnalyzedDocument(uri)?.tree;
        if (parsedTree === undefined) {
            return [];
        }
        parsedTree.rootNode.children.forEach((childNode) => {
            if (childNode.type === 'variable_assignment' && childNode.firstNamedChild?.text === 'SRC_URI') {
                TreeSitterUtils.forEach(childNode, (n) => {
                    const followChild = n.type !== 'string_content';
                    if (n.type === 'string_content') {
                        this.processSymbolsInStringContent(n, uriRegex, (start, end, match) => {
                            const matchedUri = match.groups?.uri;
                            if (matchedUri !== undefined) {
                                links.push({
                                    value: matchedUri,
                                    range: {
                                        start,
                                        end
                                    }
                                });
                            }
                            return [];
                        });
                    }
                    return followChild;
                });
            }
        });
        return links;
    }
    positionIsInRange(line, character, range) {
        return line === range.start.line && character >= range.start.character && character <= range.end.character;
    }
    calculateSymbolPositionInStringContent(n, index, match) {
        const start = {
            line: n.startPosition.row + index,
            character: match.index !== undefined ? match.index + n.startPosition.column : 0
        };
        const end = {
            line: n.startPosition.row + index,
            character: match.index !== undefined ? match.index + n.startPosition.column + match[0].length : 0
        };
        if (index > 0) {
            start.character = match.index ?? 0;
            end.character = (match.index ?? 0) + match[0].length;
        }
        return {
            start,
            end
        };
    }
    /**
     *
     * @param n The syntax node of type `string_content`
     * @param regex The regex to match the symbols
     * @param func The custom function to process the matched symbols
     */
    processSymbolsInStringContent(n, regex, func) {
        const splittedStringContent = n.text.split(/\r?\n/g);
        for (let i = 0; i < splittedStringContent.length; i++) {
            const lineText = splittedStringContent[i];
            for (const match of lineText.matchAll(regex)) {
                if (match !== undefined) {
                    const { start, end } = this.calculateSymbolPositionInStringContent(n, i, match);
                    func(start, end, match);
                }
            }
        }
    }
    /**
     *
     * @param scanResult
     * @param uri
     *
     * Process scan results sent from the client. The scan results are generated by bitbake -e <recipe> command
     */
    processRecipeScanResults(scanResult, chosenRecipe) {
        if (this.parser === undefined) {
            OutputLogger_1.logger.debug('[ProcessRecipeScanResults] The analyzer is not initialized with a parser');
            return undefined;
        }
        if (chosenRecipe === undefined) {
            OutputLogger_1.logger.error('[ProcessRecipeScanResults] The chosenRecipe is undefined, abort processing scan results');
            return;
        }
        const lines = scanResult.split(/\r?\n/g);
        const index = lines.findIndex((line) => line.includes('INCLUDE HISTORY'));
        if (index === -1) {
            OutputLogger_1.logger.debug('[ProcessRecipeScanResults] Cannot find INCLUDE HISTORY in scan results, abort processing scan results');
            return undefined;
        }
        const scanResultText = lines.slice(index).join('\r\n');
        const scanResultParsedTree = this.parser.parse(scanResultText);
        const scanResultGlobalDeclarations = (0, declarations_1.getGlobalDeclarations)({ tree: scanResultParsedTree, uri: 'scanResultDummyUri', getFinalValue: true });
        const scanResultSymbols = this.getAllSymbolsFromGlobalDeclarations(scanResultGlobalDeclarations);
        this.uriToLastScanResult[chosenRecipe] = {
            symbols: scanResultSymbols,
            includeHistory: this.extractIncludeHistory(scanResult, index)
        };
    }
    extractIncludeHistory(scanResult, includeHistoryIndex) {
        const result = [];
        const lines = scanResult.split(/\r?\n/g);
        /**
         * # INCLUDE HISTORY
         * #
         * # ...include paths we want
         * #
         * # ...operation history for variables
         */
        for (let i = includeHistoryIndex + 2; i < lines.length; i++) {
            if (lines[i] === '#') {
                break;
            }
            result.push(lines[i]);
        }
        return result.map((line) => path_1.default.parse(line.slice(1).trim()));
    }
    /**
     * @param symbol The symbol to match
     * @param lookUpSymbolList The list of symbols to look up the symbol
     * @returns The symbol that matches the symbol in the scan results
     *
     * Match the symbol in the scan results with the symbol in the original document. If the exact symbol is not found, it will try to find the symbol without 0 overrides as fallback
     */
    matchSymbol(symbol, lookUpSymbolList) {
        return lookUpSymbolList.find((scanResultDocSymbol) => {
            return this.symbolsAreTheSame(scanResultDocSymbol, symbol);
        }) ??
            lookUpSymbolList.find((scanResultDocSymbol) => {
                return scanResultDocSymbol.name === symbol.name && scanResultDocSymbol.overrides.length === 0;
            });
    }
    /**
     *
     * @param symbol A symbol that contains variable expansion syntax. e.g. VAR:${PN} or require ${PN}.inc
     * @param lookUpSymbolList A list of symbols to look up the final value of the variables in the variable expansion syntax
     *
     * Resolve the symbol if it contains variable expansion syntax in its overrides
     */
    resolveSymbol(symbol, lookUpSymbolList) {
        if (typeof symbol !== 'string') {
            const resolvedOverrides = symbol.overrides.map((override) => {
                return this.resolveVariableExpansionValues(override, lookUpSymbolList);
            });
            const resolvedSymbol = {
                ...symbol,
                overrides: [...resolvedOverrides]
            };
            return resolvedSymbol;
        }
        else {
            return this.resolveVariableExpansionValues(symbol, lookUpSymbolList);
        }
    }
    resolveVariableExpansionValues(symbol, lookUpSymbolList) {
        const regex = /\$\{(?<variable>.*)\}/g;
        let resolvedSymbol = symbol;
        for (const match of symbol.matchAll(regex)) {
            const variable = match.groups?.variable;
            if (variable !== undefined) {
                const value = lookUpSymbolList.find((symbol) => symbol.name === variable)?.finalValue;
                value !== undefined && (resolvedSymbol = resolvedSymbol.replace(new RegExp('\\$\\{' + variable + '\\}', 'g'), value.endsWith('+git') ? value.replace('+git', '') : value)); // PV usually has +git appended to it
            }
        }
        return resolvedSymbol;
    }
    extractModificationHistoryFromComments(symbol) {
        const comments = symbol.commentsAbove;
        const history = [];
        let regex = /()?/g; // dummy regex
        if (symbol.kind === vscode_languageserver_1.SymbolKind.Variable) {
            regex = /(?<=#\s{3}set\??\s)(?<filePath>\/.*):(?<lineNumber>\d+)/g;
        }
        else if (symbol.kind === vscode_languageserver_1.SymbolKind.Function) {
            regex = /(?<=#\s)line:\s(?<lineNumber>\d+),\sfile:\s(?<filePath>\/.*)/g;
        }
        comments?.forEach((comment) => {
            /**
             *  Examples:
             *  Variable:
             *  #   set /home/projects/poky/meta/conf/bitbake.conf:396
                #     [_defaultval] "gnu"
                TC_CXX_RUNTIME="gnu"
      
                Function:
                # line: 331, file: /home/projects/poky/meta/classes-global/base.bbclass
                base_do_configure(){
      
                }
             */
            for (const match of comment.matchAll(regex)) {
                const filePath = match.groups?.filePath;
                const lineNumber = match.groups?.lineNumber;
                if (filePath !== undefined && lineNumber !== undefined) {
                    const uri = 'file://' + filePath;
                    // Assuming the variables start at the beginning of the line (character 0)
                    // Could traverse the target file tree-sitter tree to find the exact range of the variable, but it costs performance
                    const location = vscode_languageserver_1.Location.create(uri, { start: { line: parseInt(lineNumber) - 1, character: 0 }, end: { line: parseInt(lineNumber) - 1, character: symbol.name.length } });
                    history.push(location);
                }
            }
        });
        return history;
    }
    /**
     *
     * @param symbolA
     * @param symbolB
     * @returns
     *
     * This functions doesn't check the deep equality of the two symbols. It only checkes the neccesary fields to determine if the two symbols are referring the same thing in the file.
     */
    symbolsAreTheSame(symbolA, symbolB) {
        const conditions = symbolA.name === symbolB.name && symbolA.overrides.length === symbolB.overrides.length;
        // Only compare the overrides that are not variable expansions.
        // For symbols with variable expansions, first resolve the values of the variables in '${}' then reconstrcut a symbol with only literal overrdies before comparing
        const overridesWithValuesForA = symbolA.overrides.filter(override => typeof override === 'string');
        const overridesWithValuesForB = symbolB.overrides.filter(override => typeof override === 'string');
        return conditions && overridesWithValuesForA.every((override) => overridesWithValuesForB.includes(override));
        // Question: should we care about the order of the overrides?
    }
}
exports.default = Analyzer;
exports.analyzer = new Analyzer();
//# sourceMappingURL=analyzer.js.map